package fi.aalto.data.server;

import java.util.Locale;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.StandardSocketOptions;
import java.time.Duration;

import com.prosysopc.ua.stack.builtintypes.LocalizedText;
import com.prosysopc.ua.stack.builtintypes.NodeId;
import com.prosysopc.ua.stack.builtintypes.UnsignedInteger;
import com.prosysopc.ua.stack.builtintypes.UnsignedShort;
import com.prosysopc.ua.stack.core.ApplicationDescription;
import com.prosysopc.ua.stack.core.Attributes;
import com.prosysopc.ua.stack.core.BrokerTransportQualityOfService;
import com.prosysopc.ua.stack.core.DataSetFieldContentMask;
import com.prosysopc.ua.stack.transport.security.AllowAllCertificatesValidator;
import com.prosysopc.ua.stack.transport.security.KeyPair;
import com.prosysopc.ua.stack.transport.security.SecurityMode;
import com.prosysopc.ua.stack.utils.EndpointUtil;
import com.prosysopc.ua.stack.cert.PkiDirectoryCertificateStore;
import com.prosysopc.ua.stack.cert.DefaultCertificateValidator;
import com.prosysopc.ua.stack.transport.security.SecurityPolicy;
import com.prosysopc.ua.stack.core.MessageSecurityMode;
import com.prosysopc.ua.stack.core.UadpDataSetMessageContentMask;
import com.prosysopc.ua.stack.core.UadpNetworkMessageContentMask;
import com.prosysopc.ua.UserTokenPolicies;
import com.prosysopc.ua.ApplicationIdentity;
import com.prosysopc.ua.UaNodeId;
import com.prosysopc.ua.UaApplication.Protocol;
//import com.prosysopc.ua.pubsub.PubSubBrokerDataSetWriterTransportSettingsConf;
import com.prosysopc.ua.pubsub.PubSubException;
import com.prosysopc.ua.pubsub.PubSubFieldTargetConf;
import com.prosysopc.ua.pubsub.PubSubJsonWriterGroupConf;
import com.prosysopc.ua.pubsub.PubSubMqttProviderConf;
import com.prosysopc.ua.pubsub.PubSubSystem;
import com.prosysopc.ua.pubsub.PubSubSystemConf;
import com.prosysopc.ua.pubsub.PubSubSystemOverrides;
import com.prosysopc.ua.pubsub.PubSubTargetVariablesConf;
import com.prosysopc.ua.pubsub.PubSubUdpUadpDataSetWriterConf;
import com.prosysopc.ua.pubsub.PubSubUdpUadpReaderGroupConf;
import com.prosysopc.ua.pubsub.PubSubUdpUadpWriterGroupConf;
import com.prosysopc.ua.pubsub.PubSubUdpUadpConnectionConf;
import com.prosysopc.ua.pubsub.PubSubUdpUadpDataSetReaderConf;
import com.prosysopc.ua.pubsub.PubSubUsernamePasswordBrokerCredentials;
import com.prosysopc.ua.pubsub.PubSubVariableDataSetConf;
import com.prosysopc.ua.pubsub.PublisherId;
import com.prosysopc.ua.server.UaServer;
import com.prosysopc.ua.server.UserValidator;

import fi.aalto.data.server.DataNodeManager;

public class DataServer {

	protected static String APP_NAME = "DataServer";
	protected static int TCP_PORT = 52580;
	protected static boolean enableSessionDiagnostics = false;
	protected DataNodeManager dataNodeManager;
	protected static UaServer server;
	
	private static final String DATA_SET_VARIABLES = "Measurements";
	private static PubSubSystem pubSubSystem;
	private static final String PUBLISHER_ID = "DataServer";
	private static final UnsignedShort VARIABLE_DATASETWRITER_ID = UnsignedShort.valueOf(432);
	private static final UnsignedShort VARIABLE_DATASETWRITER2_ID = UnsignedShort.valueOf(433);

	public static void main(String[] args) {

		DataServer dataServer = new DataServer();
		dataServer.initialize(TCP_PORT, APP_NAME);
		dataServer.createAddressSpace();
		dataServer.run(enableSessionDiagnostics);
	}
	
	public UaServer getUaServer() {
        return this.server;
	}
	
	protected void initialize(int port, String applicationName) {

		try {
		    server = new UaServer();
		    
		    ApplicationDescription appDescription = new ApplicationDescription();
		    appDescription.setApplicationName(new LocalizedText(applicationName, Locale.ENGLISH));
		    appDescription.setApplicationUri("urn:localhost:OPCUA" + applicationName);
			appDescription.setProductUri("urn:prosysopc.com:OPCUA" + applicationName);
		    server.setPort(Protocol.OpcTcp, port);
		    server.setServerName(applicationName);
		    server.setBindAddresses(EndpointUtil.getInetAddresses());

		    final PkiDirectoryCertificateStore certificateStore = new PkiDirectoryCertificateStore();
		    //final DefaultCertificateValidator validator = new DefaultCertificateValidator(certificateStore);
		    final AllowAllCertificatesValidator validator = new AllowAllCertificatesValidator();
			server.setCertificateValidator(validator);
		    File privatePath = new File(certificateStore.getBaseDir(), "private");
			KeyPair issuerCertificate = ApplicationIdentity.loadOrCreateIssuerCertificate(
					"DataCA", privatePath, "data", 3650, false);
			int[] keySizes = null;
		    final ApplicationIdentity identity = ApplicationIdentity
					.loadOrCreateCertificate(appDescription, "Data Organisation",
					"data", privatePath, null, keySizes, true);
			String hostName = ApplicationIdentity.getActualHostName();
		    server.setApplicationIdentity(identity);

		    Set<SecurityPolicy> supportedSecurityPolicies = new HashSet<SecurityPolicy>();
		    supportedSecurityPolicies.add(SecurityPolicy.NONE);
		    supportedSecurityPolicies.addAll(SecurityPolicy.ALL_SECURE_104);
		    Set<MessageSecurityMode> supportedMessageSecurityModes = new HashSet<MessageSecurityMode>();
		    supportedMessageSecurityModes.add(MessageSecurityMode.None);
		    supportedMessageSecurityModes.add(MessageSecurityMode.SignAndEncrypt);
		    server.getSecurityModes()
	        .addAll(SecurityMode.combinations(supportedMessageSecurityModes, supportedSecurityPolicies));

		    server.addUserTokenPolicy(UserTokenPolicies.ANONYMOUS);
		    server.addUserTokenPolicy(UserTokenPolicies.SECURE_USERNAME_PASSWORD);
		    //server.addUserTokenPolicy(UserTokenPolicies.SECURE_CERTIFICATE);
		    UserValidator userValidator = new DataUserValidator();
		    server.setUserValidator(userValidator);
			server.init();
			server.getSessionManager().setMaxSessionCount(500);
			server.getSessionManager().setMaxSessionTimeout(3600000);
			server.getSubscriptionManager().setMaxSubscriptionCount(500);

		} catch (Exception e) {
			System.out.println(e.getMessage());
			return;
		}

	}

	protected void createAddressSpace() {
		try {
			dataNodeManager = new DataNodeManager(server, DataNodeManager.NAMESPACE);
			dataNodeManager.getIoManager().addListeners(new DataIoManagerListener());
		} catch (Exception e) {
			System.out.println(e.getMessage());
			return;
		}
	}
	
	protected void run(boolean enableSessionDiagnostics) {
		try {
			server.start();
			if (enableSessionDiagnostics)
				server.getNodeManagerRoot().getServerData().getServerDiagnosticsNode().setEnabled(true);

			Thread t = new Thread(new DataSimulation(server));
			t.start();
			//Thread.sleep(200);
			initPubSubSystem("opc.udp://224.0.5.1:4840");
			waitForEnter();
            if (pubSubSystem != null) pubSubSystem.shutdown();
			t.stop();

			System.out.println("Closing down...");
			server.shutdown(5, new LocalizedText("Closed by user", Locale.ENGLISH));

		} catch (Exception e) {
			//System.out.println(e.getMessage());
			return;
		}
	}

	protected void waitForEnter() {
		try {
		    BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
            System.out.print("DataServer running. Press enter to exit");
            br.readLine();
		} catch (Exception e) {
			System.out.println(e.getMessage());
			return;
		}
    }
	
    public static void initPubSubSystem(String publisherAddress) {

	    PubSubSystemConf.Builder pubSubConf = PubSubSystemConf.builder();

		pubSubConf.addOrReplaceDataSet(createPubSubVariableDataSet());

		final PubSubUdpUadpWriterGroupConf uadpWriterGroupConf = createPubSubUdpUadpPublisher();
	    final PubSubUdpUadpReaderGroupConf uadpReaderGroupConf = createPubSubUadpSubscriber();
		
		pubSubConf.addOrReplaceConnection(createPubSubUdpUadpConnection(publisherAddress, uadpWriterGroupConf, uadpReaderGroupConf));
		pubSubConf.setEnabled(true);
		    
		createPubSubSystem(pubSubConf.build());
    }
    
    private static PubSubVariableDataSetConf createPubSubVariableDataSet() {
		PubSubVariableDataSetConf.Builder variableDataSetConf = PubSubVariableDataSetConf.builder();
		variableDataSetConf.addPublishedVariables(UaNodeId.fromLocal(new NodeId(2, "P300_MeasVal"), server.getNamespaceTable()));
		variableDataSetConf.setName(DATA_SET_VARIABLES);
		return variableDataSetConf.build();
    }

	private static PubSubUdpUadpWriterGroupConf createPubSubUdpUadpPublisher() {

	    PubSubUdpUadpWriterGroupConf.Builder writerGroupConf = PubSubUdpUadpWriterGroupConf.builder();

		writerGroupConf.setEnabled(true);
		writerGroupConf.setName("WriterGroup1");
		//writerGroupConf.setPublisherId(PUBLISHER_ID);
		writerGroupConf.setPublishingInterval(Duration.ofMillis(400));
		writerGroupConf.setSamplingOffset(Duration.ZERO);
		//writerGroupConf.addPublishingOffset(Duration.ofMillis(200));

		UadpNetworkMessageContentMask uadpNetworkMessageContentMask = UadpNetworkMessageContentMask.of(
		    UadpNetworkMessageContentMask.Fields.PayloadHeader, UadpNetworkMessageContentMask.Fields.SequenceNumber,
		    UadpNetworkMessageContentMask.Fields.DataSetClassId, UadpNetworkMessageContentMask.Fields.PublisherId,
		    UadpNetworkMessageContentMask.Fields.Timestamp, UadpNetworkMessageContentMask.Fields.GroupHeader,
		    UadpNetworkMessageContentMask.Fields.NetworkMessageNumber);
		writerGroupConf.setMessageContentMask(uadpNetworkMessageContentMask);

		writerGroupConf.setKeepAliveTime(Duration.ofSeconds(120));
		writerGroupConf.setWriterGroupId(UnsignedShort.valueOf(1));
		writerGroupConf.setMaxNetworkMessageSize(UnsignedInteger.valueOf(4096));

		PubSubUdpUadpDataSetWriterConf variableWriterConf = createPubSubUdpUadpVariableDataSetWriter();
		writerGroupConf.addOrReplaceWriter(variableWriterConf);


		return writerGroupConf.build();
	}

    private static PubSubUdpUadpDataSetWriterConf createPubSubUdpUadpVariableDataSetWriter() {

	    PubSubUdpUadpDataSetWriterConf.Builder variableWriterConf = PubSubUdpUadpDataSetWriterConf.builder();
		DataSetFieldContentMask fieldContentMask = DataSetFieldContentMask.of(DataSetFieldContentMask.Fields.StatusCode, DataSetFieldContentMask.Fields.SourceTimestamp);
		variableWriterConf.setFieldContentMask(fieldContentMask);

		UadpDataSetMessageContentMask msgContentMask = UadpDataSetMessageContentMask.of(UadpDataSetMessageContentMask.Fields.Timestamp);
		variableWriterConf.setMessageContentMask(msgContentMask);

		variableWriterConf.setDataSetName(DATA_SET_VARIABLES);

		variableWriterConf.setKeyFrameCount(UnsignedInteger.valueOf(5));

		variableWriterConf.setDataSetWriterId(VARIABLE_DATASETWRITER_ID);
		variableWriterConf.setEnabled(true);
		variableWriterConf.setName("MeasurementDataSetWriter");


		return variableWriterConf.build();
    }

    private static PubSubUdpUadpConnectionConf createPubSubUdpUadpConnection(String publisherAddress, PubSubUdpUadpWriterGroupConf uadpWriterGroupConf, PubSubUdpUadpReaderGroupConf uadpReaderGroupConf) {

		PubSubUdpUadpConnectionConf.Builder udpUadpConnectionConf = PubSubUdpUadpConnectionConf.builder();
		udpUadpConnectionConf.setEnabled(true);
		udpUadpConnectionConf.setName("Connection1");
		//udpUadpConnectionConf.setPublisherId(PUBLISHER_ID);
		udpUadpConnectionConf.setPublisherId(PublisherId.of(PUBLISHER_ID));

		publisherAddress = publisherAddress.replaceFirst("opc.udp://", "");
		int idx = publisherAddress.indexOf(":");
		InetSocketAddress inetAdddress = new InetSocketAddress(publisherAddress.substring(0, idx), Integer.parseInt(publisherAddress.substring(idx + 1)));
		udpUadpConnectionConf.setNetworkAddress(inetAdddress);

		NetworkInterface ni = selectNetworkInterface(true);
		udpUadpConnectionConf.putSocketOptions(StandardSocketOptions.IP_MULTICAST_IF, ni);
		udpUadpConnectionConf.putDiscoverySocketOptions(StandardSocketOptions.IP_MULTICAST_IF, ni);

		udpUadpConnectionConf.addOrReplaceGroup(uadpWriterGroupConf);
	    udpUadpConnectionConf.addOrReplaceGroup(uadpReaderGroupConf);

		return udpUadpConnectionConf.build();
	}

	static NetworkInterface selectNetworkInterface(boolean showOnlyInterfaceThatAreUp) {
		List<NetworkInterface> networkInterfaces = new ArrayList<>();
		try {
		    List<NetworkInterface> networkInterfacesTmp = Collections.list(NetworkInterface.getNetworkInterfaces());
		    for (NetworkInterface ni : networkInterfacesTmp) {
		        if (showOnlyInterfaceThatAreUp) {
		            if (ni.isUp()) {
		              networkInterfaces.add(ni);
		            }
		        } else {
		          networkInterfaces.add(ni);
		        }
		    }
		} catch (SocketException e) {
		    throw new RuntimeException("Could not get a list of available NetworkInterfaces or resolve their states", e);
		}
		if (networkInterfaces.isEmpty()) {
		    throw new RuntimeException("No NetworkInterfaces found!");
		}
		return networkInterfaces.get(0);
		  
	}
    
	private static void createPubSubSystem(PubSubSystemConf conf) {
		PubSubSystemOverrides.Builder overrides = PubSubSystemOverrides.builder();

		pubSubSystem = PubSubSystem.createFor(server, conf, overrides.build());

		try {
		    pubSubSystem.start();
		} catch (PubSubException e) {
		    System.out.print("Failed to start PubSubSystem ");
		    e.printStackTrace(System.out);
		}
	}
	
	private static PubSubUdpUadpReaderGroupConf createPubSubUadpSubscriber() {

		final PubSubUdpUadpDataSetReaderConf.Builder variableReaderConfBuilder = PubSubUdpUadpDataSetReaderConf.builder();
		
		final PubSubTargetVariablesConf.Builder targetVariablesConfBuilder = PubSubTargetVariablesConf.builder();
		targetVariablesConfBuilder.addTargetVariables(PubSubFieldTargetConf.builder().setAttributeId(Attributes.Value).setTargetNodeId(UaNodeId.from(DataNodeManager.NAMESPACE, "M200_CurCtrlVal")).build());
		variableReaderConfBuilder.setSubscribedDataSet(targetVariablesConfBuilder.build());
		variableReaderConfBuilder.setName("VariableDataSetReader");
		variableReaderConfBuilder.setEnabled(true);
		variableReaderConfBuilder.setDataSetWriterId(VARIABLE_DATASETWRITER2_ID);

		PubSubUdpUadpDataSetReaderConf variableReaderConf = variableReaderConfBuilder.build();

		PubSubUdpUadpReaderGroupConf.Builder uadpReaderConfBuilder = PubSubUdpUadpReaderGroupConf.builder();
		uadpReaderConfBuilder.addOrReplaceReader(variableReaderConf);
		uadpReaderConfBuilder.setEnabled(true);
        uadpReaderConfBuilder.setName("ReaderGroup1");
		uadpReaderConfBuilder.setMaxNetworkMessageSize(UnsignedInteger.valueOf(4096));
		
        return uadpReaderConfBuilder.build();
	}


}

